/**************************************************************/
/* sieve.c                                                    */
/* Copyright 2004, Chris Monico.                              */
/**************************************************************/
/*  This file is part of GGNFS.
*
*   GGNFS is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   GGNFS is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with GGNFS; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <gmp.h>
#include "ggnfs.h"

#define _CATCH_SIGNALS
#define DEFAULT_OUTNAME "spairs.out"
#define DEFAULT_CACHE     524288
#define DEFAULT_REPORTINTERVAL 5.0
#define SPEED_SMOOTH_B 10

/* These are (probably unreasonable) defaults! They should absolutely
   not be relied on as anything but perhaps a starting point!

  I will, at some point, compute the expected values for these, but
  I am in a hurry at the moment.
*/
#define MAX_DEFAULTS 6
s32 defaultRLIM[MAX_DEFAULTS]={0,1000,2000,  4000,  8000,   15000};
s32 defaultALIM[MAX_DEFAULTS]={0,1000,2000,  4000,  8000,   15000};
s32 defaultMPR[MAX_DEFAULTS]={0,  0,  0,20000, 20000, 500000};
s32 defaultMPA[MAX_DEFAULTS]={0,  0,  0,20000, 20000, 500000};
int  defaultLP[MAX_DEFAULTS] ={0,  0,  1,    1,     1,      2};
int  defaultLPA[MAX_DEFAULTS]={0,  0,  1,    1,     1,      2};
double defaultRLambda[MAX_DEFAULTS]={1.0,1.1, 1.3, 1.4,    1.5,     1.6};
double defaultALambda[MAX_DEFAULTS]={1.0,1.1, 1.3, 1.4,    1.5,     1.6};
#define DEFAULT_FB_NAME "default.fb"

#define USAGE \
"[OPTIONS]\n"\
"--help              : show this help and exit\n"\
"-fb <filename>      : use factor base in file <filename>\n"\
"-j  <filename>      : use job file <filename>\n"\
"-test               : Experimental! (requires -fb and -j)"

#define START_MSG \
"\n"\
" __________________________________________________________ \n"\
"|         This is the sieve program for GGNFS.             |\n"\
"| Version: %-15s                                 |\n"\
"| This program is copyright 2004, Chris Monico, and subject|\n"\
"| to the terms of the GNU General Public License version 2.|\n"\
"|__________________________________________________________|\n"


/**** Globals ****/
double startTime, reportInterval = DEFAULT_REPORTINTERVAL;
extern int forceStop;
int    isTest=0;
double rTested=0, rSmooth=0;
double aTested=0, aSmooth=0;
double rSmoothPerc, aSmoothPerc;

int procPairs_cl(char *fname, relation_t *Rels, int numRels, nfs_fb_t *FB);

/**************************************************/
void catch_term(int sig)
/**************************************************/
{ 
  printf("Term signal caught. Doing clean shutdown...");
  fflush(stdout);
  forceStop=1;
}

/*******************************************************/
void setDependentDefaults(nfs_sieve_job_t *job)
/*******************************************************/
/* `job' has at least a number to be factored and the  */
/* `nDigits' field filled in. This function will set   */
/* any other fields which haven't yet been initialized */
/* to some default values depending on `nDigits'.      */
/*******************************************************/
{ int n, w=0;
  char warn[]="** Warning: using default value for %s!";

  n = job->nDigits/10;
  n = MIN(n, MAX_DEFAULTS-1);

  if ((job->FB.rfb_size == -1)&&(job->FB.rLim == -1)) {
    job->FB.rLim = defaultRLIM[n]; w++;
    msgLog("", warn, "rLim");
  }
  if ((job->FB.afb_size == -1)&&(job->FB.aLim == -1)) {
    job->FB.aLim = defaultALIM[n]; w++;
    msgLog("", warn, "aLim");
  }
  if (job->FB.maxP_r == -1) {
    job->FB.maxP_r = defaultMPR[n]; w++;
    msgLog("", warn, "mpr");
  }
  if (job->FB.maxP_a == -1) {
    job->FB.maxP_a = defaultMPA[n]; w++;
    msgLog("", warn, "mpa");
  }
  if (job->FB.maxLP == -1)
    job->FB.maxLP = defaultLP[n];
  if (job->FB.maxLPA == -1)
    job->FB.maxLPA = defaultLPA[n];

  if (job->FB.rfb_lambda < 0 ) {
    job->FB.rfb_lambda = job->rfb_lambda = defaultRLambda[n]; w++;
    msgLog("", warn, "rlambda");
  }
  if (job->FB.afb_lambda < 0) {
    job->FB.afb_lambda = job->afb_lambda = defaultALambda[n]; w++;
    msgLog("", warn, "alambda");
  }
  if (w) 
    msgLog("", "** Warning: Unreliable default values used for %d important fields!", w);

}

/*******************************************************/
int parseJobFile(nfs_sieve_job_t *job, char *fName)
/*******************************************************/
{ FILE *fp;
  char token[256], value[256], thisLine[1024];
  int  t;

  if (!(fp = fopen(fName, "r"))) {
    printf("Error opening %s for read!\n", fName);
    return -1;
  }
  job->nDigits=0;
  /* Quick check to see if the polynomial is being specified in the job file: */
  while (!(feof(fp))) {
    thisLine[0]=0;
    fgets(thisLine, 1023, fp);
    if (strncmp(thisLine, "START_POLY", 10)==0) {
      printf("Found polynomial in job file. Reading...\n");
      readPoly(fp, &job->FB);
      job->nDigits = mpz_sizeinbase(job->FB.n, 10);
    }
  }
  rewind(fp); /* Back up to the beginning of the file. */
    
  
  /* Initialize fields to empty, or defaults which are independent of
     the number being factored. */
  job->type = SIEVE_UNKNOWN;
  job->a0 = job->a1 = job->b0 = job->b1 = 0;
  job->qIndex0 = job->qIndex1 = 0;
  job->sieveArea = 0.0;
  job->FB.rfb_size = job->FB.afb_size = -1;
  job->FB.rLim = job->FB.aLim = -1;
  job->FB.maxP_r = job->FB.maxP_a = -1;
  job->FB.maxLP = job->FB.maxLPA = -1;
  job->cacheSize=DEFAULT_CACHE;
  job->rfb_lambda = job->afb_lambda = -1;
  strcpy(job->outName, DEFAULT_OUTNAME);

  
  while (!(feof(fp))) {
    thisLine[0] = 0;
    fgets(thisLine, 1023, fp);
    /* Special case: If there's a polynomial, skip it: */
    if (strncmp(thisLine, "START_POLY", 10)==0) {
      while (!(feof(fp)) && (strncmp(thisLine, "END_POLY", 8)))
        fgets(thisLine, 1023, fp);
      thisLine[0]='#'; thisLine[1]=0; /* Dirty hack so the loop below is ok. */
    }
    if ((sscanf(thisLine, "%255s %255s", token, value)==2) && (thisLine[0] != '#')) {
      if (strncmp(token, "type:", 5)==0) {
        if (strncmp(value, "classical", 9)==0)
          job->type = SIEVE_CLASSICAL;
        else if (strncmp(value, "lattice", 7)==0)
          job->type = SIEVE_LATTICE;
        else
          job->type = SIEVE_UNKNOWN;
      } else if (strncmp(token, "a0:",3)==0) {
        job->a0 = atol(value);
      } else if (strncmp(token, "a1:",3)==0) {
        job->a1 = atol(value);
      } else if (strncmp(token, "b0:",3)==0) {
        job->b0 = atol(value);
      } else if (strncmp(token, "b1:",3)==0) {
        job->b1 = atol(value);
      } else if (strncmp(token, "q0:",3)==0) {
        job->qIndex0 = atol(value);
      } else if (strncmp(token, "q1:",3)==0) {
        job->qIndex1 = atol(value);
      } else if (strncmp(token, "sieveArea:",10)==0) {
        job->sieveArea = atof(value);
      } else if (strncmp(token, "cacheSize:",3)==0) {
        job->cacheSize = atol(value);
      } else if (strncmp(token, "rlambda:",4)==0) {
        job->rfb_lambda = atof(value);
      } else if (strncmp(token, "alambda:",4)==0) {
        job->afb_lambda = atof(value);
      } else if (strncmp(token, "fbname:", 7)==0) {
        strncpy(job->fbName, value, MAXFNAMESIZE);
      } else if (strncmp(token, "rfb_size:", 9)==0) {
        job->FB.rfb_size = atol(value);
      } else if (strncmp(token, "afb_size:", 9)==0) {
        job->FB.afb_size = atol(value);
      } else if (strncmp(token, "rLim:", 5)==0) {
        job->FB.rLim = atol(value);
      } else if (strncmp(token, "aLim:", 5)==0) {
        job->FB.aLim = atol(value);
      } else if (strncmp(token, "mpr:", 4)==0) {
        t = atoi(value); t = MIN(t, 31);
        job->FB.maxP_r = (1<<t);
      } else if (strncmp(token, "mpa:", 4)==0) {
        t = atoi(value); t = MIN(t, 31);
        job->FB.maxP_a = (1<<t);
      } else if (strncmp(token, "maxLP:", 6)==0) {
        job->FB.maxLP = atoi(value);
      } else if (strncmp(token, "maxLPA:", 6)==0) {
        job->FB.maxLPA = atoi(value);
     }
    }
  }
  fclose(fp);
  if (job->type == SIEVE_UNKNOWN) {
    printf("Couldn't determine sieve type!\n");
    return -1;
  }
  return 0;
}

/***********************************************************/     
int c1Sieve(nfs_sieve_job_t *J)
{ s32   maxSmooth, numSmooth, totSmooth, *smooth_a, lastAdjust;
  s32   i, a0, a1, b0, b1, b, candidates, cacheSize, totCand=0;
  s32   recentPairs[SPEED_SMOOTH_B], recentTotal;
  double recentTimes[SPEED_SMOOTH_B], recentTime, thisTime;
  relation_t *R;
  time_t now, start;
  double tStart, tNow, lastReport, smoothPerc;
  nfs_fb_t *FB=&J->FB;
  FILE *fp;

  /* Set up parameters. */  
  FB->rfb_lambda = J->rfb_lambda;
  FB->afb_lambda = J->afb_lambda;
  a0 = J->a0;
  a1 = J->a1;
  b0 = J->b0;
  b1 = J->b1;
  cacheSize = J->cacheSize;  
  if ((fp = fopen(J->outName, "a"))) {
    fprintf(fp, "# GGNFS version %s classical siever output.\n", GGNFS_VERSION);
    fprintf(fp, "# number: %s\n", FB->name);
    fclose(fp);
  }

  for (i=0; i<SPEED_SMOOTH_B; i++) {
    recentPairs[i]=0;
    recentTimes[i]=0;
  }  
  recentTotal = 0; recentTime=0;

  maxSmooth = MAX_PAIRS_PER_CALL;
  smooth_a = (s32 *)malloc(maxSmooth*sizeof(s32));
  R = (relation_t *)malloc(maxSmooth*sizeof(relation_t));

  printf("Doing classical sieve on a=[%ld, %ld] from b=%ld to b=%ld.\n",
          a0, a1, b0, b1);
  printf("Large rational prime bound : %ld\n", FB->maxP_r);
  printf("Large algebraic prime bound: %ld\n", FB->maxP_a);
  lastReport = sTime()-100.0;  
  

  totSmooth=0;
  time(&start);
  lastAdjust = 0;
  tStart = sTime();
  tNow = sTime();
  /*************************/
  /* This is the main loop */
  /*************************/
  for (b=b0 ; (b<=b1) && !(forceStop); b++) {
    thisTime = tNow; 
    candidates = clCacheFriendlySieve(smooth_a, maxSmooth, FB, a0, a1, b, cacheSize);
    if (candidates >= MAX_PAIRS_PER_CALL-1) {
      fprintf(stderr, "Warning: MAX_PAIRS_PER_CALL=%ld reached! Some relations were likely missed!\n",
        (s32)MAX_PAIRS_PER_CALL);
    }
    for (i=0; i<candidates; i++) {
      R[i].a = smooth_a[i]; R[i].b = b;
    }
    numSmooth = procPairs_cl(J->outName, R, candidates, FB);
    tNow = sTime();
    thisTime = tNow - thisTime;
    totSmooth += numSmooth;
    totCand += candidates;

    recentTotal = recentTotal - recentPairs[b%SPEED_SMOOTH_B] + numSmooth;
    recentTime = recentTime - recentTimes[b%SPEED_SMOOTH_B] + thisTime;
    recentPairs[b%SPEED_SMOOTH_B] = numSmooth;  recentTimes[b%SPEED_SMOOTH_B] = thisTime;

    if (tNow > (lastReport+reportInterval)) {
      if (recentTotal > 0) {
        smoothPerc = 100.0*(double)totSmooth/(double)totCand;
        printf("TR: %ld, b=%ld, newR:%ld, %1.3lf%% GRep (%1.2lf%%/%1.2lf%%), sec/r: %1.6lf\n", 
                totSmooth, b, numSmooth, smoothPerc, rSmoothPerc, aSmoothPerc,
                recentTime/recentTotal);
      } else printf("TR: 0, b=%ld, newR:0, 0%% GRep\n", b);
      lastReport = tNow;
    }

  }
  time(&now);
  printf("Found %ld smooth pairs in total\n", totSmooth);
  printf("Elapsed time: %ld seconds.\n", now-start);

  msgLog("", "clSieved: a=%ld..%ld, b=%ld..%ld",
         a0, a1, b0, b-1);

  free(smooth_a); free(R);
  return totSmooth;
}

#define B_SAMPLE_SIZE 4
/***********************************************************/     
int c1SieveTest(nfs_sieve_job_t *J, int numPieces)
/**************************************************************/
/* Estimate the number of good relations which would be found */
/* if we sieved over the rectangle specified in 'J'.          */
/* `numPieces' is the number of tiny regions which will be    */
/* sieved to get an extrapolation.                            */
/**************************************************************/
{ s32   maxSmooth, numSmooth, *smooth_a, numCands;
  s32   i, j, a0, a1, b0, b1, b, candidates, cacheSize;
  s32   lastB0, thisB0, thisB1;
  double tStart, lastAvgSmooth=0.0, avgSmoothPerB, estTotSmooth;
  double thisTime, lastTime=-1.0, estTotTime=0.0, numS, numT;
  double multiplier, lastStartB;
  nfs_fb_t *FB=&J->FB;

  /* Set up parameters. */  
  FB->rfb_lambda = J->rfb_lambda;
  FB->afb_lambda = J->afb_lambda;
  a0 = J->a0;
  a1 = J->a1;
  b0 = J->b0;
  b1 = J->b1;
  cacheSize = J->cacheSize;  
  
  maxSmooth = 100000;
  smooth_a = (s32 *)malloc(maxSmooth*sizeof(s32));
  

  printf("Doing classical sieve TEST on a=[%ld, %ld], b=[%ld, %ld]...\n",
          a0, a1, b0, b1);
//  printf("Large rational prime bound : %ld\n", FB->maxP_r);
//  printf("Large algebraic prime bound: %ld\n", FB->maxP_a);

  estTotSmooth=0;
  /*************************/
  /* This is the main loop */
  /*************************/
  lastB0=-1;  

  multiplier = exp((1/(double)numPieces)*log((double)b1/(double)b0));
  lastStartB = (double)b0/multiplier;
  for (j=0; j<numPieces; j++) {
    lastStartB *= multiplier;
    thisB0 = (s32)lastStartB;
    /* To make sampling as uniform as possible, make sure
       we always start with an odd 'b' and end with an even. */
    if (thisB0%2==0) thisB0++;   
    thisB1 = thisB0 + B_SAMPLE_SIZE-1;
    if (thisB1%2==1) thisB1++;
    numSmooth = 0;
    thisTime = 0.0;
    numS = numT = 0.0;
    numCands = 0;
    for (b=thisB0 ; (b<=thisB1) && !(forceStop); b++) {
      tStart = sTime();
      candidates = clSieve(smooth_a, maxSmooth, FB, a0, a1, b, cacheSize);
      numCands += candidates;
      thisTime += sTime()-tStart;
      for (i=0; i<candidates; i++) {
        if (rand()%10==0) {
          numT += 1.0;
          if ((isSmooth_alg(smooth_a[i], b, FB) >=0) && (isSmooth_rat(smooth_a[i], b, FB)>=0))
            numS += 1.0; 
        }
      }
    }
    if (numT > 0.0)
      numSmooth = (s32)((double)numCands*numS/numT);
    else numSmooth = 0;
    avgSmoothPerB = (double)numSmooth/(double)(thisB1-thisB0+1);
    thisTime /= B_SAMPLE_SIZE;
    if (lastB0 > 0) {
      /* Estimate the number of relations that would be found in 
         the interval [lastB0, thisB0).
         This is definitely NOT a very accurate estimate! We will
         do just a linear extrapolation. In the future, we might
         consider using a more accurate model. Probably logarithmic
         decay would be closer, but still not right.
      */
      estTotSmooth += (thisB0 - lastB0)*(avgSmoothPerB + lastAvgSmooth)/2.0;

      /* Estimate the time to do the given interval. For this, linear
         is ok. */
      estTotTime += (thisB0 - lastB0)*(thisTime + lastTime)/2.0;
    }
    lastB0 = thisB0;
    lastAvgSmooth = avgSmoothPerB;
    lastTime = thisTime;
    printTmp("%1.0lf%% completed...", (double)100.0*(1.0+j)/numPieces);
  }
  
  printf("Estimate %ld relations, %1.3lf hours.\n", (s32)estTotSmooth, estTotTime/3600.0);

  free(smooth_a);
  return estTotSmooth;
}
/*************************************************************/
int clSieveEstimate(nfs_sieve_job_t *J)
{ double area, aSize, bSize, skew;
  int    i;
  nfs_fb_t *FB=&J->FB;

  area = 10000000;
  skew = FB->skew;
  if (skew < 0.001) 
    /* Assign something arbitrary. */
    skew = 20.0; 
  for (i=0; i<7; i++) {
    bSize = sqrt(area/skew);
    aSize = bSize*skew;
    J->a0 = -(s32)aSize/2;
    J->a1 = -J->a0;
    J->b0 = 10;
    J->b1 = bSize+10;
    printf("*********************************\n");
    printf("area = %e.\n", area);
    c1SieveTest(J, 20);
    area *= 10.0;
  }
  return 0;
}
#ifdef _NO
/***********************************************************/     
int l1Sieve(nfs_sieve_job_t *J)
{ s32   maxSmooth, numSmooth, totSmooth, candidates, *cands;
  s32   i, qIndex, q0, q1, a, b, cacheSize=DEFAULT_CACHE;
  s32   rFails=0, aFails=0;
  time_t now, start;
  double tStart, tNow, lastReport, area=1000000;
  int    itsSinceCandTest=0;
  double smoothPerc, weightedAvg=0, weight=0;
  nfs_fb_t *FB=&J->FB;
  FILE   *ofp;

  /* Set up parameters. */  
  FB->rfb_lambda = J->rfb_lambda;
  FB->afb_lambda = J->afb_lambda;
  area = J->sieveArea;
  cacheSize = J->cacheSize;
  q0 = J->qIndex0;
  q1 = J->qIndex1;
  
  maxSmooth = 5000;
  cands = (s32 *)malloc(2*maxSmooth*sizeof(s32));
  

  printf("Doing lattice sieve with sieveArea=%1.0lf, cacheSize=%ld\n", area, cacheSize);
  printf("Large rational prime bound : %ld\n", FB->maxP_r);
  printf("Large algebraic prime bound: %ld\n", FB->maxP_a);
  
  
  lastReport = sTime()-100.0;  
  totSmooth=0;
  time(&start);
  tStart = sTime();

  /*************************/
  /* This is the main loop */
  /*************************/
  numSmooth = 0;
  for (qIndex=q0 ; (qIndex<q1) && !(forceStop); qIndex++) {
    candidates = latSieve(cands, maxSmooth, FB, qIndex, area, cacheSize);
    numSmooth = 0;

    tNow = sTime();
    ofp = fopen(J->outName, "a");
    for (i=0; i<candidates; i++) {
      fprintf(ofp, "%ld %ld %ld\n", cands[2*i], cands[2*i+1], qIndex);
    }
    fclose(ofp);
    totSmooth += candidates;
    time(&now);

    tNow = sTime();
    if ((tNow > (lastReport+5.0))&&(totSmooth > 0)) {
      if ((itsSinceCandTest++)%5==0) {
        numSmooth=0;
        for (i=0; i<candidates; i++) {
          a = cands[2*i]; b = cands[2*i+1];
          if (isSmooth_alg(a, b, FB) >=0) {
            if (isSmooth_rat(a, b, FB)>=0) {
              numSmooth++;
            } else rFails++;
          } else aFails++;
//          if ((isSmooth_alg(a, b, FB) >=0) && (isSmooth_rat(a, b, FB)>=0))
//            numSmooth++;
        }
        smoothPerc = ((candidates>0) ? (double)100.0*numSmooth/(double)candidates : 100.0);
        weightedAvg = (weight*weightedAvg + candidates*smoothPerc)/(weight + candidates);
        weight += candidates;
        weightedAvg = MAX(weightedAvg, 0.0000001);
        printf("total: %ld, qIndex = %ld, newR: %ld (%1.3lf%% are smooth, avg=%1.3lf), sec/rel: %1.6lf\n", 
                totSmooth, qIndex, candidates, smoothPerc, weightedAvg, 
                (tNow - tStart)/(totSmooth*weightedAvg/100.0));
        printf("    aFails=%ld, rFails=%ld\n", aFails, rFails);
      } else {
        printf("total: %ld, qIndex = %ld, newR: %ld, sec/rel: %1.6lf\n", 
                totSmooth, qIndex, candidates, (tNow - tStart)/(totSmooth*weightedAvg/100.0));
      }
      lastReport = tNow;
    }

  }
  time(&now);
  printf("Found %ld smooth pairs in total\n", totSmooth);
  printf("Elapsed time: %ld seconds.\n", now-start);
  msgLog("", "latSieved: q=%ld..%ld", q0, qIndex-1);
  free(cands);
  return totSmooth;
}
#endif

/*****************************************************/
int procPairs_cl(char *fname, relation_t *Rels, int numRels, nfs_fb_t *FB)
/*****************************************************/
/* The (a,b) candidates are stored in Rels in just   */
/* The 'a' and 'b' fields. This function will factor */
/* them, remove bad ones, and dump the good ones to  */
/* the given file in the format expected by procrels.*/
/* Return value: The number of good relations.       */
/*****************************************************/
{ int  i, j, numGood=0;
  char str[2048];
  FILE *fp;

  numGood = isSmooth_rat_withInfo_par(Rels, numRels, FB);
  if (forceStop) return 0;
  rTested += numRels; rSmooth += numGood;
  rSmoothPerc = 100.0*rSmooth/(1.0 + rTested);
  for (i=j=0; i<numRels; i++) {
    if (Rels[i].b>0) {
      memcpy(&Rels[j], &Rels[i], sizeof(relation_t));
      j++;
    }
  }
  numRels = j;
  numGood = isSmooth_alg_withInfo_par(Rels, numRels, FB);
  if (forceStop) return 0;
  aTested += numRels; aSmooth += numGood;
  aSmoothPerc = 100.0*aSmooth/(1.0 + aTested);
  if (!(fp = fopen(fname, "a"))) {
    fprintf(stderr, "Error opening %s for append!!!\n", fname);
    return numGood;
  }
  for (i=0; i<numRels; i++) {
    if (Rels[i].b>0) {
      makeOutputLine(str, &Rels[i], FB);
      fprintf(fp, "%s\n", str);
    }
  }
  fclose(fp);
  return numGood;
}

/***************************************************/
int tryMakeFB(nfs_sieve_job_t *J)
/***************************************************/
/* Try to create a factor base from the info in J. */
/* Return 0 on success, nonzero on failure.        */
/***************************************************/
{ int res;

  /* setDependentDefaults() will try to fill in missing fields with
     (probably bad!) default values. It will log warnings
     when defaults values are used in important fields.
  */
  setDependentDefaults(J);

  res = createFB(&J->FB, J->fbName);
  return res;
}

/**************************************************/
int main(int argC, char *args[])
{ char            fbname[MAXFNAMESIZE], jobfile[MAXFNAMESIZE];
  int             i, res, fbLoaded=0;
  nfs_sieve_job_t Job;
  double          now;
  
  fbname[0]=jobfile[0]=0;
  forceStop=0;
  strncpy(Job.fbName, DEFAULT_FB_NAME, MAXFNAMESIZE);

  printf(START_MSG, GGNFS_VERSION);

  /* Register signal handler for clean shutdowns. */
#ifdef _CATCH_SIGNALS
  signal(SIGTERM, catch_term);
  signal(SIGINT, catch_term);
#ifndef _MSC_VER
  signal(SIGQUIT, catch_term);
#endif
#endif
  
  /***************************/
  /* Parse command-line args */
  /***************************/
  for (i=1; i<argC; i++) {
    if (strcmp(args[i], "-fb")==0) {
      if ((++i)<argC) {
        strcpy(fbname, args[i]);
      }
    }
    else if (strcmp(args[i], "-j")==0) {
      if ((++i)<argC) {
        strcpy(jobfile, args[i]);
      }
    }
    else if (strcmp(args[i], "-test")==0) {
      isTest=1;
    }
    else if (strcmp(args[i], "--help")==0) {
      printf("USAGE: %s %s\n", args[0], USAGE);
      return 0;
    }
  }
  if (!(jobfile[0])) {
    printf("USAGE: %s %s\n", args[0], USAGE);
    return 0;
  }
  if (!isTest)  
    msgLog("", "GGNFS-%s : sieve", GGNFS_VERSION);
  else {
    printf("Warning: test-mode is very very crude - the results\n"
           "are potentially very far from true (particularly with\n"
           "unreasonable parameters!\n");
  }
  /***********************************/
  /* Initialize and get factor base. */
  /***********************************/
  initFB(&Job.FB);


  
  /******************/
  /* Get parameters */
  /******************/
  if (parseJobFile(&Job, jobfile)) {
    printf("Some error occured parsing job file. Cannot continue.\n");
    exit(-1);
  }
  if (loadFB(fbname, &Job.FB)) {
    fbLoaded=0;
    fprintf(stderr, "Could not load factor base from %s. Attempting to create...\n", 
            fbname);
    if (tryMakeFB(&Job)) {
      fprintf(stderr, "Factor base creation failed! Aborting.\n");
      exit(-1);
    }
  }
  setLogs(&Job.FB);
  msgLog("", "Log bases: %1.3lf, %1.3lf.", Job.FB.rfb_log_base, Job.FB.afb_log_base);



  startTime = sTime();
  if (Job.type == SIEVE_CLASSICAL) {
    if (!isTest)
      res = c1Sieve(&Job);
    else 
      clSieveEstimate(&Job);
  } else if (Job.type == SIEVE_LATTICE) {
    printf("Lattice sieving not implemented for this program!\n");
    exit(-1);
  }

  now = sTime();
  if (!isTest)
    msgLog("", "SieveTime: %1.1lf", now-startTime);
  
  return 0;
}




